/*  $Id: idl.pl,v 1.1.1.1 2001/01/24 15:20:59 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(idl,
	  [ idl_file_to_term/2,		% +File, -Term
	    test/1,
	    test/0
	  ]).
:- use_module(idldcg).		% First-phase parsing
:- use_module(util).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This module calls the first-phase   parser  and subsequently reorganises
the parsetree and resolves scoped names,   making  the parstree suitable
for further processing.

The structure of the resulting term is   described below. Many slots are
lists of things. In this case, the  list   is  named  in plural, and the
description of the members of the list singular. Thus, <Methods> means a
Prolog list of <Method> terms.

The current parser deals with the complete   OMG IDL Grammar, except for
interface inheritance, ::  references  of   scoped  names  and operation
contexts.

<Term>		::= <Definition>
<Definition>	::= definition(<Dcl>, <Interfaces>, <Modules>)
<Dcl>		::= dcls(<Consts>, <Types>, <Exceptions>)

<Const>		::= const(<Name>, <Scope>, <Value>)

<Type>		::= type(<Name>, <Scopes>, <TypeSpec>)
<TypeSpec>	::= scalar(<ScalarType>)
		  | char
		  | octet
		  | boolean
		  | any
		  | string(<SeqSize>)
		  | sequence(<Type>, <SeqSize>)
		  | struct(<StructMembers>)
		  | union(<Type>, <Cases>)
		  | enum(<EnumMembers>)
<Scope>		::= <ModuleName>	% last in innermost scope
		  | <InterfaceName>
<ScalarType>	::= float
		  | double
		  | long
		  | short
		  | unsigned(long)
		  | unsigned(short)
<SeqSize>	::= <integer>
		  | unbound
<StructMember>	::= <Name>:<Type>
<Case>		::= case(<Value>, <Type>, <Name>)
		  | case(@(<Name>, <Index>), <Type>, <Name>)	% enum
<EnumMember>	::= <Atom>		  
<Index>		::= <1..>

<Exception>	::= exception(<Name>, <Scope>, <StructMembers>)

<Interface>	::= interface(<Name>, <Scope>, <Dcl>, <Attributes>, <Methods>)

<Attribute>	::= attribute(<Name>, <Type>, <Access>)
<Access>	::= get			% readonly
		  | both

<Method>	::= method(<Name>, <Parms>, <Return>, <Exceptions>, <Context>)
<Parm>		::= in(<Type>, <Name>)
		  | out(<Type>, <Name>)
		  | inout(<Type>, <Name>)
<Return>	::= <Type>

<Context>	::= ??? (TBD)

<Module>	::= module(<Name>, <Definition>)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


idl_file_to_term(File, Term) :-
	progress(parse_idl(File, List)),
	progress(resolve_def(List, [], dcls([], [], []), Term)).

resolve_def(List, Scope, DclsI, Def) :-
	Def0 = definition(Const0,
			  Types0,
			  Except0,
			  Interfaces0,
			  Modules0),
	Def  = definition(dcls(Const, Types, Except),
			  Interfaces,
			  Modules),
	reorder_def(List, Def0),
	make_interface_types(Interfaces0, Scope, ItfTypes),
	merge_dcls(DclsI, dcls([], ItfTypes, []), Dcls0),
	resolve_consts(Const0, Scope, Dcls0, Const),
	merge_dcls(Dcls0, dcls(Const, [], []), Dcls1),
	resolve_types(Types0, Scope, Dcls1, Types),
	merge_dcls(Dcls1, dcls([], Types, []), Dcls2),
	resolve_exceptions(Except0, Scope, Dcls2, Except),
	merge_dcls(Dcls2, dcls([], [], Except), Dcls3),
	resolve_interfaces(Interfaces0, Scope, Dcls3, Interfaces),
	resolve_modules(Modules0, Scope, Dcls3, Modules).


reorder_def([], Term) :-
	make_all_args_nil(Term).
reorder_def([H|T], Tin) :-
	classify_def(H, N),
	replace_arg(N, Tin, [H|R], Tout, R),
	reorder_def(T, Tout).

classify_def(H, N) :-
	functor(H, Name, _),
	classify_def_from_functor(Name, N).

classify_def_from_functor(const,     1).
classify_def_from_functor(typedef,   2).
classify_def_from_functor(struct,    2).
classify_def_from_functor(enum,      2).
classify_def_from_functor(union,     2).
classify_def_from_functor(exception, 3).
classify_def_from_functor(interface, 4).
classify_def_from_functor(module,    5).

classify_def_from_functor(attribute, 4). % for interface handling
classify_def_from_functor(method,    5).

		 /*******************************
		 *	   DECLARATIONS		*
		 *******************************/

%	dcl(Consts, Types): Declaration context in effect

merge_dcls(dcls(Const0, Types0, Except0),
	   dcls(Const1, Types1, Except1),
	   dcls(Const, Types, Except)) :- !,
	append(Const1, Const0, Const),
	append(Types1, Types0, Types),
	append(Except1, Except0, Except).
merge_dcls(dcls(Consts, Types, Except),
	   Const,
	   dcls([Const|Consts], Types, Except)) :-
	dcls_arg_of(Const, 1), !.
merge_dcls(dcls(Consts, Types, Except),
	   Type,
	   dcls(Consts, [Type|Types], Except)) :-
	dcls_arg_of(Type, 2), !.
merge_dcls(dcls(Consts, Types, Excepts),
	   Except,
	   dcls(Consts, Types, [Except|Excepts])) :-
	dcls_arg_of(Except, 3).

in_dcls(Term, Dcls) :-
	dcls_arg_of(Term, N),
	arg(N, Dcls, Arg),
	memberchk(Term, Arg).

dcls_arg_of(const(_, _, _),	1).
dcls_arg_of(type(_, _, _),	2).
dcls_arg_of(exception(_, _, _),	3).


		 /*******************************
		 *	       MODULES		*
		 *******************************/

resolve_modules([], _, _, []).
resolve_modules([module(Name, Def0)|T0], Scope0, Dcls,
		[module(Name, Def1)|T1]) :-
	append(Scope0, [Name], Scope),
	resolve_def(Def0, Scope, Dcls, Def1),
	resolve_modules(T0, Scope0, Dcls, T1).



		 /*******************************
		 *	        CONST		*
		 *******************************/

%	resolve_consts(+RawConst, +Dcls, +Bindings)
%
%	Resolve all consts declarations.  After this, consts are of the
%	format `const(Name, Scope, Value)', where Name is an atom, and Value is
%	an atom, integer or float.  `Context' are consts from the current
%	scope.
%
%	TBD: if the type is not resolved, it cannot be validated, but we
%	cannot do the types first, as types may use consts for lengths of
%	sequences and arrays.

resolve_consts(In, Scope, Dcls, Out) :-
	resolve_consts(In, Scope, Dcls, UnRes, Out0),
	(   UnRes == []
	->  Out = Out0
	;   In == UnRes
	->  error(unresolved(UnRes))
	;   merge_dcls(Dcls, dcls(Out0, [], []), Dcls1),
	    resolve_consts(UnRes, Scope, Dcls1, Out1),
	    append(Out0, Out1, Out)
	).

resolve_consts([], _, _, [], []).
resolve_consts([const(Name, Type, V0)|CT0], Scope, Dcls, UnRes, [Const|CT]) :-
	resolve_const(Type, V0, Dcls, Value), !,
	Const = const(Name, Scope, Value),
	merge_dcls(Dcls, Const, Dcls1),
	resolve_consts(CT0, Scope, Dcls1, UnRes, CT).
resolve_consts([C0|CT0], Scope, Dcls, [C0|UnRes], CT) :-
	resolve_consts(CT0, Scope, Dcls, UnRes, CT).

resolve_const(Type, V0, Dcls, V2) :-
	eval_expr(V0, Dcls, V1),
	(   satifies(V1, Type, V2)
	->  true
	;   error(bad_const_value(V0))
	).

satifies(V, scalar(long), V)    :- !, integer(V).
satifies(I, scalar(short), V)   :- !, integer(I), V is (I << 16) >> 16.
satifies(V, scalar(float), V)   :- !, float(V).
satifies(V, scalar(double), V)  :- !, float(V).
satifies(V, string(unbound), V) :- !, atom(V).
satifies(V, string(N), V)       :- !, atom(V), atom_length(V, L), L =< N.
satifies(V, _, V).				% just assume ok for now

eval_expr(scoped(Name), Dcls, V) :- !,
	in_dcls(const(Name, V), Dcls).
eval_expr(X, _Dcls, X) :-
	atomic(X).
eval_expr(+X0, Dcls, X) :- !,
	eval_expr(X0, Dcls, X),
	number(X).
eval_expr(-X0, Dcls, X) :- !,
	eval_expr(X0, Dcls, X1),
	number(X1),
	X is -X1.
eval_expr(Term, Dcls, X) :-
	functor(Term, Name, 2), !,
	arg(1, Term, Left0),
	arg(2, Term, Right0),
	eval_expr(Left0, Dcls, Left),
	eval_expr(Right0, Dcls, Right),
	number(Left),
	number(Right),
	Expr =.. [Name, Left, Right],
	X is Expr.
	
	
		 /*******************************
		 *	       TYPEDEF		*
		 *******************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Resolve types to  primitive  CORBA  types,   removing  all  scoped  name
references. The result is a list of types of the format:

	type(Name, Scope, Simple)
	type(Name, Scope, array(Type, [Dimensions]))
	type(Name, Scope, sequence(Type, UnboundOrSize))
	type(Name, Scope, struct([Name:Type]))
	type(Name, Scope, union(SwitchType, Members))
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


resolve_types(In, Scope, Dcls, Out) :-
	resolve_types5(In, Scope, Dcls, UnRes, Out0),
	(   UnRes == []
	->  Out = Out0
	;   In == UnRes
	->  error(unresolved(UnRes))
	;   merge_dcls(Dcls, dcls([], Out0, []), Dcls1),
	    resolve_types(UnRes, Scope, Dcls1, Out1),
	    append(Out0, Out1, Out)
	).


resolve_types5([], _, _, [], []).
resolve_types5([H0|T0], Scope, Dcls, UnRes, Types) :-
	resolve_type(H0, Scope, Dcls, H1), !,
	merge_dcls(Dcls, dcls([], H1, []), Dcls1),
	append(H1, T1, Types),
	resolve_types5(T0, Scope, Dcls1, UnRes, T1).
resolve_types5([H0|T0], Scope, Dcls, [H0|UnRes], T1) :-
	resolve_types5(T0, Scope, Dcls, UnRes, T1).


resolve_type(typedef(Names0, typedef(Names1, Type0)), Scope, Dcls, List) :- !,
	expand_type(Type0, Scope, Dcls, Type1),
	append(Names0, Names1, Names2),
	list_to_set(Names2, Names),
	iterate_types(Names, Type1, Scope, Dcls, List).
resolve_type(typedef(Names, Type0), Scope, Dcls, List) :-
	expand_type(Type0, Scope, Dcls, Type1),
	iterate_types(Names, Type1, Scope, Dcls, List).
	
iterate_types([], _, _, _, []).
iterate_types([array(Name, Dim0)|Names],
	      Type, Scope, Dcls,
	      [type(Name, array(Type, Dim))|Types]) :- !,
	expand_array_dims(Dim0, Dcls, Dim),
	iterate_types(Names, Type, Scope, Dcls, Types).
iterate_types([Name|Names], Type, Scope, Dcls,
	      [type(Name, Scope, Type)|Types]) :-
	iterate_types(Names, Type, Scope, Dcls, Types).

expand_type(Type0, Scope, Dcls, Type) :-
	do_expand_type(Type0, Scope, Dcls, Type), !.
expand_type(Type0, _, _, _) :-
	Type0 \= scoped(_),		% might be done later
	error(unresolved_type(Type0)).
	
do_expand_type(scoped(Name), _Scope, Dcls, Type) :- !,
	Type0 = type(Name, Scope, Def),
	in_dcls(Type0, Dcls),
	(   Def = type(_Name2, _Scope2, Def2)
	->  Type = type(Name, Scope, Def2)
	;   Type = Type0
	).
do_expand_type(string(scoped(Name)), _Scope, Dcls, string(N)) :- !,
	in_dcls(const(Name, _, N), Dcls),
	integer(N), N > 0.
do_expand_type(sequence(Type0, Len), Scope, Dcls, sequence(Type, Len)) :- !,
	expand_type(Type0, Scope, Dcls, Type).
do_expand_type(struct(Members0), Scope, Dcls, struct(Members)) :- !,
	expand_struct_member_types(Members0, Scope, Dcls, Members).
do_expand_type(union(Switch0, Cases0), Scope, Dcls, union(Switch, Cases)) :- !,
	expand_type(Switch0, Scope, Dcls, Switch),
	(   Switch = type(_, _, enum(Enum))
	->  expand_union_enum_cases(Cases0, Enum, Scope, Dcls, Cases)
	;   expand_union_cases(Cases0, Scope, Dcls, Cases)
	).
do_expand_type(Type, _, _, Type).
	
%	expand_named_type(+Type0, +Scope, +Dcls, -Type)
%
%	Expand a type.  If the type is compound and unnamed, give it
%	an anonymous name.  These names are <Scope>_AnonType_<N>_, as defined
%	by the ILU interface (ISL cannot handle in-lined type definition).
%
%	How is the numbering actually done?  We assume lexically.

expand_named_type(Type0, Scope, Dcls, Type) :-
	expand_type(Type0, Scope, Dcls, Type1),
	(   setting(orb, ilu),
	    \+ Type1 = type(_Name, _Scope, _Def),
	    compound_type(Type1)
	->  anon_number(Type1, TypeN),
	    concat_atom(['AnonType', TypeN, ''], '_', AnonType),
	    Type = type(AnonType, Scope, Type1)
	;   Type = Type1
	).

compound_type(sequence(_, _)).
compound_type(struct(_)).
compound_type(union(_, _)).

anon_number(_, N) :-
	flag(anon_type, X, X+1),
	N is X + 1.
%:- dynamic
%	anon_type/2.
%anon_number(Type, N) :-
%	anon_type(Type, N), !.
%anon_number(Type, N) :-
%	(   anon_type(_, Prev)
%	->  N is Prev + 1
%	;   N = 1
%	),
%	asserta(anon_type(Type, N)).


%	expand_array_dims(+Dims0, +Dcls, -Dims)
%
%	Deal with the definition of array dimensions using const
%	declarations.

expand_array_dims([], _, []).
expand_array_dims([scoped(Name)|T0], Dcls, [N|T]) :- !,
	in_dcls(const(Name, N), Dcls),
	integer(N), N > 0,
	expand_array_dims(T0, Dcls, T).
expand_array_dims([N|T0], Consts, [N|T]) :-
	expand_array_dims(T0, Consts, T).

expand_struct_member_types([], _, _, []).
expand_struct_member_types([Name:Type0|T0], Scope, Dcls, [Name:Type|T]) :-
	expand_named_type(Type0, Scope, Dcls, Type),
	expand_struct_member_types(T0, Scope, Dcls, T).

expand_union_enum_cases([], _, _, _, []).
expand_union_enum_cases([H0|T0], Enum, Scope, Dcls, [H|T]) :-
	expand_union_enum_case(H0, Enum, Scope, Dcls, H),
	expand_union_enum_cases(T0, Enum, Scope, Dcls, T).

expand_union_enum_case(case(scoped(EnumName), Type0, Name), Enum, Scope, Dcls,
		       case(@(EnumName, N), Type, Name)) :- !,
	(   nth1(N, Enum, EnumName)
	->  expand_named_type(Type0, Scope, Dcls, Type)
	;   error(no_value_in_enum(EnumName))
	).
expand_union_enum_case(case(Val, _, _), _, _, _, _) :-
	error(no_value_in_enum(Val)).
	
expand_union_cases([], _, _, []).
expand_union_cases([H0|T0], Scope, Dcls, [H|T]) :-
	expand_union_case(H0, Scope, Dcls, H),
	expand_union_cases(T0, Scope, Dcls, T).

expand_union_case(case(Val0, Type0, Name), Scope, Dcls,
		 case(Val,  Type,  Name)) :-
	expand_named_type(Type0, Scope, Dcls, Type),
	(   Val0 = scoped(ConstName)
	->  in_dcls(const(ConstName, _, Val), Dcls)
	;   Val = Val0
	).

make_interface_types([], _, []).
make_interface_types([interface(Itf, _Export)|T0],
		     Scope,
		     [type(Itf, Scope, interface)|T]) :-
	make_interface_types(T0, Scope, T).


		 /*******************************
		 *	     EXCEPTIONS		*
		 *******************************/

resolve_exceptions([], _, _, []).
resolve_exceptions([exception(Name, Members0)|T0], Scope, Dcls,
		   [exception(Name, Scope, Members)|T]) :-
	resolve_exception_members(Members0, Scope, Dcls, Members),
	resolve_exceptions(T0, Scope, Dcls, T).

resolve_exception_members([], _, _, []).
resolve_exception_members([Name:Type0|T0], Scope, Dcls, [Name:Type|T]) :-
	expand_type(Type0, Scope, Dcls, Type), !,
	resolve_exception_members(T0, Scope, Dcls, T).
resolve_exception_members([Name:Type0|T0], Scope, Dcls, [Name:Type0|T]) :-
	error(unresolved_type(exception_member(Name), Type0)),
	resolve_exception_members(T0, Scope, Dcls, T).


		 /*******************************
		 *	     INTERFACES		*
		 *******************************/

resolve_interfaces([], _, _, []).
resolve_interfaces([I0|T0], Scope, Dcls, [I|T]) :-
	resolve_interface(I0, Scope, Dcls, I),
	resolve_interfaces(T0, Scope, Dcls, T).

resolve_interface(interface(Itf, Exports), Scope0, Dcls0,
		  interface(Itf,
			    Scope0,
			    dcls(Consts, Types, Except),
			    Atts, Methods)) :-
	Def = definition(Consts0,
			 Types0,
			 Except0,
			 Atts0,
			 Methods0),
	reorder_def(Exports, Def),
	append(Scope0, [Itf], Scope),
	resolve_consts(Consts0, Scope, Dcls0, Consts),
	merge_dcls(Dcls0, dcls(Consts, [], []), Dcls1),
	resolve_types(Types0, Scope, Dcls1, Types),
	merge_dcls(Dcls1, dcls([], Types, []), Dcls2),
	resolve_exceptions(Except0, Scope, Dcls2, Except),
	merge_dcls(Dcls1, dcls([], [], Except), Dcls3),
	resolve_attributes(Atts0, Dcls3, Atts),
	resolve_methods(Methods0, Dcls3, Methods).

		 /*******************************
		 *	    ATTRIBUTES		*
		 *******************************/

resolve_attributes([], _, []).
resolve_attributes([attribute(Names, Type0, Access)|T0], Dcls, Atts) :-
	expand_type(Type0, [], Dcls, Type), 	% TBD: pass in scope
	iterate_atts(Names, Type, Access, Atts0),
	append(Atts0, T, Atts),
	resolve_attributes(T0, Dcls, T).

iterate_atts([], _, _, []).
iterate_atts([Name|Names], Type, Access, [attribute(Name, Type, Access)|T]) :-
	iterate_atts(Names, Type, Access, T).


		 /*******************************
		 *	      METHODS		*
		 *******************************/

resolve_methods([], _, []).
resolve_methods([method(Name, Parms0, Return0, Raises0, Ctx)|T0], Dcls,
		[method(Name, Parms, Return, Raises, Ctx)|T]) :-
	resolve_parameters(Parms0, Dcls, Parms),
	expand_type(Return0, [], Dcls, Return), % TBD: pass in scope
	expand_raises(Raises0, Dcls, Raises),
	resolve_methods(T0, Dcls, T).

resolve_parameters([], _, []).
resolve_parameters([P0|T0], Dcls, [P|T]) :-
	resolve_parameter(P0, Dcls, P),
	resolve_parameters(T0, Dcls, T).

resolve_parameter(in(Type0, Name), Dcls, in(Type, Name)) :-
	expand_type(Type0, [], Dcls, Type).	% TBD: pass in scope
resolve_parameter(out(Type0, Name), Dcls, out(Type, Name)) :-
	expand_type(Type0, [], Dcls, Type).
resolve_parameter(inout(Type0, Name), Dcls, inout(Type, Name)) :-
	expand_type(Type0, [], Dcls, Type).

expand_raises([], _, []).
expand_raises([scoped(Except)|T0], Dcls, [Exception|T]) :-
	Exception = exception(Except, _, _),
	(   in_dcls(Exception, Dcls)
	->  true
	;   error(no_exception(Except))
	),
	expand_raises(T0, Dcls, T).


		 /*******************************
		 *	        UTIL		*
		 *******************************/

%	replace_arg(+N, +Tin, +Ain, -Tout, -Aout)

replace_arg(N, Tin, Ain, Tout, Aout) :-
	functor(Tin, Name, Arity),
	functor(Tout, Name, Arity),
	replace_arg(1, N, Tin, Ain, Tout, Aout).

replace_arg(A, A, Tin, Ain, Tout, Aout) :- !,
	arg(A, Tin, Ain),
	arg(A, Tout, Aout),
	NN is A + 1,
	replace_arg(NN, A, Tin, Ain, Tout, Aout).
replace_arg(N, A, Tin, Ain, Tout, Aout) :-
	arg(N, Tin, Tmp), !,
	arg(N, Tout, Tmp),
	NN is N + 1,
	replace_arg(NN, A, Tin, Ain, Tout, Aout).
replace_arg(_, _, _, _, _, _).

%	make_all_args_nil(+Term)

make_all_args_nil(Term) :-
	functor(Term, _, Arity),
	make_all_args_nil(Arity, Term).

make_all_args_nil(0, _) :- !.
make_all_args_nil(N, Term) :-
	arg(N, Term, []), !,
	NN is N - 1,
	make_all_args_nil(NN, Term).


		 /*******************************
		 *		TEST		*
		 *******************************/

:- use_module(pretty_print).

test :-
	test(test).

test(Base) :-
	file_name_extension(Base, idl, File),
	idl_file_to_term(File, T),
	pretty_print(T).
