/*  $Id: idldcg.pl,v 1.2 2001/01/31 10:07:41 jan Exp $

    Designed and implemented by Jan Wielemaker
    E-mail: jan@swi.psy.uva.nl

    Copyright (C) 1997 University of Amsterdam. All rights reserved.
*/

:- module(idl_dcg,
	  [ parse_idl/2		% +File, -Term
	  ]).
:- use_module(util).

:- dynamic
	errors/1.			% # errors detected

parse_idl(File, Term) :-
	tmp_file(idl, Temp),
	run_cpp(File, Temp),
	open(Temp, read, In),
	get_code(In, C0),
	read_file(C0, In, FileChars),
	close(In),
	delete_file(Temp),
	retractall(errors(_)),
	asserta(errors(0)),
	phrase(idl(Term), FileChars),
	(   errors(0)
	->  true
	;   format(user_error,
		   '[idl2pl: Syntax errors detected, aborted]~n', []),
	    fail
	).

run_cpp(In, Out) :-
	current_prolog_flag(windows, true), !,
	prolog_to_os_filename(In, In1),
	prolog_to_os_filename(Out, Out1),
	sformat(CP1, 'cmd /c copy ~w tmp.cpp', [In1]),
	shell(CP1),
	sformat(Cmd, 'cl.exe -P tmp.cpp"', []),
	shell(Cmd),
	sformat(CP2, 'cmd /c copy tmp.i ~w', [Out1]),
	shell(CP2), !.
run_cpp(In, Out) :-
%	sformat(Cmd, 'sh -c "cpp -lang-c++ ~w > ~w"', [In, Out]),
	sformat(Cmd, 'sh -c "cpp -x c++ ~w > ~w"', [In, Out]), % gcc 2.95
	shell(Cmd), !.
run_cpp(In, Out) :-
	error(run_cpp(In, Out)),
	fail.


read_file(-1, _, []) :- !.
read_file(C, Fd, [C|T]) :-
	get_code(Fd, C1),
	read_file(C1, Fd, T).
	

		 /*******************************
		 *	      TOPLEVEL		*
		 *******************************/

idl(Def) -->
	cpp_line,
	blanks,
	definitions(Def),
	blanks.


		 /*******************************
		 *	     DEFINITION		*
		 *******************************/

definitions([H|T]) -->
	blanks,
	definition(H), !,
	colon,
	definitions(T).
definitions([]) --> [].

definition(Def) --> type_dcl(Def), !.
definition(Def) --> const_dcl(Def), !.
definition(Def) --> except_dcl(Def), !.
definition(Def) --> interface(Def), !.
definition(Def) --> module(Def), !.


		 /*******************************
		 *	      MODULE		*
		 *******************************/

module(module(Name, Definitions)) -->
	"module", identifier(Name),
	begin,
	    definitions(Definitions),
	end.

		 /*******************************
		 *	    INTERFACE		*
		 *******************************/

interface(interface(Name, Exports)) -->
	"interface",
	identifier(Name),
	begin, !,
	    exports(Exports),
	end.
interface(forwards(interface(Name))) -->
	"interface",
	identifier(Name).

exports([H|T]) -->
	export(H), !, colon,
	exports(T).
exports([]) -->
	[].

export(Exp) -->	type_dcl(Exp), !.
export(Exp) -->	const_dcl(Exp), !.
export(Exp) -->	except_dcl(Exp), !.
export(Exp) -->	attr_dcl(Exp), !.
export(Exp) -->	op_dcl(Exp), !.
export(_)   --> end, !, {fail}.
export(_)   --> error(syntax('')).

		 /*******************************
		 *	       METHODS		*
		 *******************************/

op_dcl(method(Name, Arguments, Return, Raises, '<context>')) -->
	op_type_spec(Return),
	identifier(Name),
	open_brace,
	(   close_brace
	->  {Arguments = []}
	;   param_dcls(Arguments),
	    close_brace
	),
	raises_expr(Raises), !.
	
op_type_spec(void) -->
	blanks,
	"void", !.
op_type_spec(Type) -->
	param_type_spec(Type).

param_dcls([H|T]) -->
	blanks,
	param_dcl(H),
	(   comma
	->  param_dcls(T)
	;   {T=[]}
	).
	
param_dcl(in(Type, Name)) -->
	"in", !,
	blanks,
	param_type_spec(Type),
	identifier(Name).
param_dcl(out(Type, Name)) -->
	"out", !,
	blanks,
	param_type_spec(Type),
	identifier(Name).
param_dcl(inout(Type, Name)) -->
	"inout", !,
	blanks,
	param_type_spec(Type),
	identifier(Name).

param_type_spec(Type) -->
	base_type_spec(Type), !.
param_type_spec(Type) -->
	string_type(Type), !.
param_type_spec(Type) -->
	scoped_name(Type).

scoped_name(scoped(Name)) -->
	identifier(Name),		% TBD: ::<identifier>, etc.
	{ \+ reserved(Name)
	}.

reserved(enum).
reserved(struct).
reserved(union).
reserved(long).
reserved(short).
reserved(string).
reserved(char).
reserved(boolean).
reserved(unsigned).
reserved(case).
reserved(default).
reserved(switch).

scoped_names([H|T]) -->
	scoped_name(H), !,
	(   comma
	->  scoped_names(T)
	;   {T = []}
	).
scoped_names([]) -->
	[].

		 /*******************************
		 *	    ATTRIBUTES		*
		 *******************************/

attr_dcl(attribute(Names, Type, Access)) -->
	attr_access(Access),
	"attribute", !,
	(   param_type_spec(Type),
	    simple_declarator_list(Names)
	->  {true}
	;   error(syntax(attribute))
	).

attr_access(get) -->
	"readonly", !,
	blanks.
attr_access(both) -->
	[].

simple_declarator_list([H|T]) -->
	simple_declarator(H),
	(   comma
	->  simple_declarator_list(T)
	;   {T = []}
	).


		 /*******************************
		 *	       TYPE		*
		 *******************************/

type_dcl(Type) -->
	"typedef", !,
	(   type_declarator(Type)
	->  {true}
	;   error(syntax(typedef))
	).
type_dcl(Type) -->
	constr_type_spec(Type).

type_declarator(typedef(Names, Type)) -->
	blanks,
	type_spec(Type),
	declarators(Names).

type_spec(Type) -->
	simple_type_spec(Type), !.
type_spec(Type) -->
	constr_type_spec(Type).

simple_type_spec(Type) -->
	base_type_spec(Type), !.
simple_type_spec(Type) -->
	template_type_spec(Type), !.
simple_type_spec(Type) -->
	scoped_name(Type).

constr_type_spec(Type) -->
	struct_type(Type).
constr_type_spec(Type) -->
	union_type(Type).
constr_type_spec(Type) -->
	enum_type(Type).

struct_type(typedef([Name], struct(Members))) -->
	"struct", !,
	(   identifier(Name),
	    begin,
	        members(Members),
	    end
	->  {true}
	;   error(syntax(struct))
	).

	
%	STRUCTS

members(Members) -->
	blanks,
	member(H), colon, !,
	members(T),
	{ append(H, T, Members)
	}.
members([]) -->
	[].

member(List) -->
	type_spec(Type),
	declarators(Names),
	{findall(Name:Type, member(Name, Names), List)}.

%	DECLARACTORS

declarators([H|T]) -->
	declarator(H),
	(   comma
	->  declarators(T)
	;   {T=[]}
	).

declarator(H) -->
	complex_declarator(H), !.
declarator(H) -->
	simple_declarator(H), !.

simple_declarator(H) -->
	identifier(H).

complex_declarator(H) -->
	array_declarator(H).

array_declarator(array(Name, Dimensions)) -->
	identifier(Name),
	fixed_array_size_list(Dimensions).

fixed_array_size_list([H|T]) -->
	fixed_array_size(H),
	(   peek("[")
	->  fixed_array_size_list(T)
	;   {T=[]}
	).

fixed_array_size(N) -->
	"[",
	positive_int_const(N),
	"]",
	blanks.
	    
template_type_spec(Template) -->
	sequence_type(Template), !.
template_type_spec(String) -->
	string_type(String).


sequence_type(sequence(Type, N)) -->
	identifier(sequence),
	lt,
	simple_type_spec(Type),
	(   comma
	->  positive_int_const(N)
	;   {N = unbound}
	),
	gt.


string_type(string(Size)) -->
	identifier(string),
	(   lt
	->  positive_int_const(Size),
	    gt
	;   {Size = unbound}
	).


base_type_spec(Type) -->
	blanks,
	(   scalar_type(Type)
	;   char_type(Type)
	;   any_type(Type)
	;   boolean_type(Type)
	;   object_type(Type)
	), !,
	blanks.
	
scalar_type(Scalar) -->
	scalar_type(Scalar, _).

scalar_type(scalar(Type), Which) -->
	identifier(Type),
	{ scalar_type(Type, Which)
	}, !.
scalar_type(scalar(unsigned(Type)), integer) -->
	"unsigned",
	blanks,
	identifier(Type),
	{ unsigned_scalar_type(Type)
	}, !.

scalar_type(float,	float).
scalar_type(double,	float).
scalar_type(octet,	octet).
scalar_type(long,	integer).
scalar_type(short,	integer).

unsigned_scalar_type(long).
unsigned_scalar_type(short).

any_type(any)	      --> "any", !.
boolean_type(boolean) --> "boolean", !.
char_type(char)       --> "char", !.
object_type('Object') --> "Object", !.

integer_type(Type) -->
	scalar_type(Type, integer).

%	UNION

union_type(typedef([Name], union(SwitchType, Body))) -->
	"union",
	(   identifier(Name),
	    "switch", open_brace, switch_type_spec(SwitchType), close_brace,
	    begin,
	        switch_body(Body),
	    end
	->  {true}
	;   error(syntax(union))
	).
	   
switch_type_spec(Type) --> integer_type(Type), !.
switch_type_spec(Type) --> char_type(Type), !.
switch_type_spec(Type) --> boolean_type(Type), !.
switch_type_spec(Type) --> enum_type(Type), !.
switch_type_spec(Type) --> scoped_name(Type).

switch_body([H|T]) -->
	case(H), !,  colon,
	switch_body(T).
switch_body([]) -->
	[].
	
case(case(Labels, Type, Name)) -->
	case_labels(Labels),
	blanks,
	type_spec(Type),
	declarator(Name), !.
case(_) -->
	\+ peek("}"),
	error(syntax('')).

case_labels(Val) -->
	"case", !,
	blanks,
	const_exp(Val),
	blanks,
	":".
case_labels(*) -->
	"default", blanks, ":".


%	ENUM

enum_type(typedef([Name], enum(Enumerators))) -->
	"enum", !,
	(   identifier(Name),
	    begin,
	        enumerators(Enumerators),
	    end
	->  {true}
	;   error(syntax(enume))
	).

enumerators([H|T]) -->
	enumerator(H),
	(   comma
	->  enumerators(T)
	;   {T=[]}
	).

enumerator(H) -->
	identifier(H).


		 /*******************************
		 *	       EXCEPT		*
		 *******************************/

except_dcl(exception(Name, Members)) -->
	"exception", !,
	(   identifier(Name),
	    begin,
	        members(Members),
	    end
	->  {true}
	;   error(syntax(exception))
	).

raises_expr(List) -->
	"raises", !,
	(   open_brace,
	        scoped_names(List),
	    close_brace
	->  {true}
	;   error(syntax(raises))
	).
raises_expr([]) -->
	[].

		 /*******************************
		 *	       CONST		*
		 *******************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Const declarations are used to share   constants  between the server and
the client. The code below parses an   OMG IDL `const_dcl' into a Prolog
term of the format:

	const(Name, Type, Value)

If `Value' is scalar, it is translated into a term ready to be evaluated
by  is/2,  except  for  possible  references   to  other  constants.  As
references may be forwards, this needs to be resolved in a second phase.
The Prolog operators used are:

	\/, /\, <<, >>, \, xor
	+. -, *, /, mod
	+, - (unary)

Note that the official OMG IDL grammar for consts is wrong, making 3+4*5
illegal syntax!  This syntax is rewritten in the spirit of this BNF, but
restoring symetry.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

const_dcl(const(Name, Type, Value)) -->
	"const", !,
	blanks,
	(   const_type(Type),
	    identifier(Name),
	    "=",
	    blanks,
	    const_exp(Value),
	    blanks,
	    peek(";")
	->  {true}
	;   error(syntax(const))
	).

const_type(Type) -->
	scalar_type(Type).
const_type(Type) -->
	string_type(Type).
const_type(Type) -->
	scoped_name(Type).

const_exp(Value) -->
	or_exp(Value).

or_exp(Value) -->
	xor_exp(Left),
	(   {Value = Left}
	;   blanks,
	    "|", !,
	    blanks,
	    xor_exp(Right),
	    { Value = Left \/ Right
	    }
	).

xor_exp(Value) -->
	and_exp(Left),
	(   {Value = Left}
	;   blanks,
	    "^", !,
	    blanks,
	    and_exp(Right),
	    { Value = xor(Left, Right)
	    }
	).

and_exp(Value) -->
	shift_exp(Left),
	(   {Value = Left}
	;   blanks,
	    "&", !,
	    blanks,
	    shift_exp(Right),
	    { Value = (Left /\ Right)
	    }
	).

shift_exp(Value) -->
	add_exp(Left),
	(   {Value = Left}
	;   blanks,
	    shift_op(Op), !,
	    blanks,
	    add_exp(Right),
	    { Value =.. [Op, Left, Right]
	    }
	).

add_exp(Value) -->
	mult_exp(Left),
	(   {Value = Left}
	;   blanks,
	    add_op(Op), !,
	    blanks,
	    mult_exp(Right),
	    { Value =.. [Op, Left, Right]
	    }
	).

mult_exp(Value) -->
	unary_exp(Left),
	(   {Value = Left}
	;   blanks,
	    mult_op(Op), !,
	    blanks,
	    unary_exp(Right),
	    { Value =.. [Op, Left, Right]
	    }
	).

shift_op(>>) --> ">>".
shift_op(<<) --> "<<".

add_op(+)    --> "+".
add_op(-)    --> "-".

mult_op(*)   --> "*".
mult_op(/)   --> "/".
mult_op(mod) --> "%".

unary_exp(+Value) --> lit("+"), !, primary_expr(Value).
unary_exp(-Value) --> lit("-"), !, primary_expr(Value).
unary_exp(\Value) --> lit("~"), !, primary_expr(Value).
unary_exp(Value)  --> primary_expr(Value).

primary_expr(Value) --> open_brace, !, const_exp(Value), close_brace.
primary_expr(Value) --> scoped_name(Value), !.
primary_expr(Value) --> literal(Value), !.

literal(Value) --> character_literal(Value).
literal(Value) --> string_literal(Value).
literal(Value) --> boolean_literal(Value).
literal(Value) --> floating_pt_literal(Value).
literal(Value) --> integer_literal(Value).

character_literal(C) -->
	"'", [C], "'", {C \== 0'\}, !.
character_literal(C) -->
	"'", character(C), "'", !.

character(C) -->			% \xhh
	"\\x",
	hex_digit(D1),
	hex_digit(D2), !,
	{ C is 16*D1 + D2 }.
character(C) -->			% \ooo
	"\\",
	octal_digit(D1),
	octal_digit(D2),
	octal_digit(D3), !,
	{ C is 64*D1 + 8*D2 + D3 }.
character(7)  --> "\\a".
character(8)  --> "\\b".
character(9)  --> "\\t".
character(10) --> "\\n".
character(12) --> "\\f".
character(13) --> "\\r".
character(92) --> "\\\\".
character(34) --> "\\\"".
character(39) --> "\\'".

hex_digit(D) --> digit(C), {D is C - 0'0}.
hex_digit(D) --> [C], {between(0'a, 0'f, C), D is C + 10 - 0'a}.
hex_digit(D) --> [C], {between(0'A, 0'F, C), D is C + 10 - 0'A}.

octal_digit(D) --> [C], {between(0'0, 0'7, C), D is C - 0'0}.

string_literal(S) -->
	"""", !, string_chars(Chars),
	{atom_codes(S, Chars)}.

string_chars(L) -->
	"""", !,
	blanks,
	(   """"
	->  string_chars(L)
	;   {L=[]}
	).
string_chars([C|T]) -->
	peek("\\"),
	character(C), !,
	string_chars(T).
string_chars([C|T]) -->
	[C],
	string_chars(T).
	
boolean_literal(true)  --> "TRUE".
boolean_literal(false) --> "FALSE".

integer_literal(Value) -->
	"0", !,
	digits(CT),
	{ number_chars(Value, [0'0,0'o|CT])
	}.
integer_literal(Value) -->
	digit(C0),
	radus(C1),
	digits(CT),
	{ number_chars(Value, [C0,C1|CT])
	}, !.
integer_literal(Value) -->
	digit(C0),
	digits(CT),
	{ number_chars(Value, [C0|CT])
	}, !.

radus(C) --> [C], {atom_char(N, C), radus_char(N)}.

radus_char(x).				% hexadecimal
radus_char('X').
radus_char('b').			% binary
radus_char('B').
radus_char('o').			% octal
radus_char('O').

floating_pt_literal(Value) -->
	digit(C0),
	digits(C1),
	(   "."
	->  digits(C2a),
	    {C2 = [0'.| C2a]}
	;   {C2 = []}
	),
	(   exponent_char(EC)
	->  optional_sign(S),
	    digit(E0),
	    digits(E1),
	    { flatten([EC, S, E0, E1], E)
	    }
	;   { E = []
	    }
	),
	{ flatten([C0, C1, C2, E], Chars),
	  number_chars(Value, Chars)
	}, !.

exponent_char(0'e) --> "e".
exponent_char(0'E) --> "E".

optional_sign(0'-) --> "-".
optional_sign(0'+) --> "+".
optional_sign([])  --> [].

positive_int_const(N) -->
	blanks,
	const_exp(N),		% TBD: Check for scoped names, etc.
	blanks.


		 /*******************************
		 *	       UTIL		*
		 *******************************/

blanks -->
	nl, !,
	cpp_line,
	blanks.
blanks -->
	[C], {C =< 32}, !,
	blanks.
blanks -->
	[].

cpp_line -->
	"#", chars, nl, !.	% skip CPP line info
cpp_line -->
	[].

begin       --> blanks, "{", blanks.
end         --> blanks, "}", blanks.
lt          -->	blanks, "<", blanks.
gt	    --> blanks, ">", blanks.
open_brace  --> blanks, "(", blanks.
close_brace --> blanks, ")", blanks.
colon	    --> blanks, ";", blanks.
comma       --> blanks, ",", blanks.

lit([S0|S]) -->
	blanks, [S0|S], blanks.

identifier(Id) -->
	{ atom(Id), !,
	  atom_codes(Id, Chars)
	},
	string(Chars),
	peek([C]),
	{ \+ alnum(C)
	}.
identifier(Id) -->
	blanks,
	letter(C0),
	alnums(CT),
	{ atom_codes(Id, [C0|CT])
	},
	blanks.

string([]) --> [].
string([H|T]) --> [H], string(T).

digits([D|T]) -->
	digit(D), !,
	digits(T).
digits([]) -->
	[].


alnums([H|T]) -->
	alnum(H), !,
	alnums(T).
alnums([]) -->
	[].


letter(C) -->
	[C],
	{ between(0'a, 0'z, C)
	; between(0'A, 0'Z, C)
	; C == 0'_
	}.

alnum(C) -->
	[C],
	{ alnum(C)
	}.

alnum(C) :-
	between(0'a, 0'z, C), !.
alnum(C) :-
	between(0'A, 0'Z, C), !.
alnum(C) :-
	between(0'0, 0'9, C), !.
alnum(0'_).


digit(C) -->
	[C],
	{ between(0'0, 0'9, C)
	}.

nl -->
	[10].

chars -->
	[].
chars -->
	[_],
	chars.

peek([C], L, L) :-
	L = [C|_].
	
		 /*******************************
		 *	   ERROR HANDLING	*
		 *******************************/

error(What) -->
	{ retract(errors(N)),
	  NN is N + 1,
	  assert(errors(NN))
	},
	[C0|C],
	peek(";"), !,
	{ print_error(What, [C0|C])
	}.

print_error(syntax(Statement), S) :-
	format(user_error,
	       '[idl2pl: Syntax error in "~w ~s;"]~n', [Statement, S]).


		 /*******************************
		 *	       PORTRAY		*
		 *******************************/

:- asserta((user:portray(List) :- idl_dcg:portray_string(List))).

portray_string(List) :-
	ascii_list(List), !,
	length(List, L),
	(   L > 40
	->  length(L1, 40),
	    append(L1, _, List),
	    write('"'),
	    write_string(L1),
	    write(' ..."')
	;   write('"'),
	    write_string(List),
	    write('"')
	).

write_string([]).
write_string([10|T]) :- !,
	write('^J'),
	write_string(T).
write_string([H|T]) :-
	put_code(H),
	write_string(T).

ascii_list([]).
ascii_list([H|T]) :-
	printable(H), !,
	ascii_list(T).

printable(9).
printable(10).
printable(13).
printable(C) :-
	integer(C),
	between(32, 126, C).
	
	
	
